
# coding: utf-8

# In[ ]:


import argparse
import numpy as np
import pandas as pd
import os
import random
import time

import matplotlib.pyplot as plt
get_ipython().run_line_magic('matplotlib', 'inline')


# In[ ]:


import torch
import torchvision
import torchvision.transforms as transforms

import torch.nn as nn
import torch.nn.functional as func
import torch.optim as optim
import torch.backends.cudnn as cudnn

from torch.optim.optimizer import required
from torch.autograd import Variable
from torch.autograd import Function


# In[ ]:


parser = argparse.ArgumentParser()
parser.add_argument('--dataset', default='mnist', help='mnist | cifar10 | svhn')
parser.add_argument('--dataroot', default='data', help='path to dataset')
parser.add_argument('--workers', type=int, default=2, help='number of data loading workers')
parser.add_argument('--batchSize', type=int, default=1024, help='input batch size')
parser.add_argument('--imageSize', type=int, default=28, help='the height / width of the input image to network')
parser.add_argument('--nClasses', type=int, default=10, help='number of labels (classes)')
parser.add_argument('--conditional', type=bool, default=True, help='condition GAN on class labels')
parser.add_argument('--nChannels', type=int, default=1, help='number of colour channels')
parser.add_argument('--nz', type=int, default=100, help='size of the latent z vector')
parser.add_argument('--ngf', type=int, default=64)
parser.add_argument('--ndf', type=int, default=64)
parser.add_argument('--filterSize', type=int, default=5)
parser.add_argument('--niter', type=int, default=100, help='number of epochs to train for')
parser.add_argument('--nCriticIter', type=int, default=5, help='number of critic iterations')
parser.add_argument('--nGeneratorIter', type=int, default=1, help='number of generator iterations')
parser.add_argument('--lr', type=float, default=0.001, help='learning rate, default=0.0002')
parser.add_argument('--beta1', type=float, default=0.5, help='beta1 for adam. default=0.5')
parser.add_argument('--beta2', type=float, default=0.9, help='beta2 for adam. default=0.9')
parser.add_argument('--lambdaaa', type=float, default=10, help='noise variance, default=0.5')
parser.add_argument('--C', type=float, default=1.0, help='embedding L2-norm bound, default=1.0')
parser.add_argument('--sigma', type=float, default=0.1, help='noise variance, default=0.5')
parser.add_argument('--nUnrolls', type=int, default=0, help='number of epochs to train for')
parser.add_argument('--cuda', action='store_true', help='enables cuda')
parser.add_argument('--ngpu', type=int, default=1, help='number of GPUs to use')
parser.add_argument('--netG', default='', help="path to netG (to continue training)")
parser.add_argument('--netD', default='', help="path to netD (to continue training)")
parser.add_argument('--outf', default='output_mnist_labeller', help='folder to output images and model checkpoints')
parser.add_argument('--manualSeed', type=int, help='manual seed')

opt, unknown = parser.parse_known_args()

try:
    os.makedirs(opt.outf)
except OSError:
    pass

if torch.cuda.is_available():
    opt.cuda = True
    opt.ngpu = 1 #torch.cuda.device_count()
    gpu_id = 0
    
if opt.manualSeed is None:
    opt.manualSeed = random.randint(1, 10000)
print("Random Seed: ", opt.manualSeed)
random.seed(opt.manualSeed)
torch.manual_seed(opt.manualSeed)
if opt.cuda:
    torch.cuda.manual_seed_all(opt.manualSeed)

cudnn.benchmark = True


# In[ ]:


ngpu = int(opt.ngpu)
nz = int(opt.nz)
ngf = int(opt.ngf)
ndf = int(opt.ndf)
nc = int(opt.nChannels)

one = torch.ones(opt.batchSize)


# In[ ]:


def onehot(labels, batch_size, n_classes):
    if not opt.conditional:
        return labels.float()
    
    idx = labels.data.contiguous().view(-1, 1) if isinstance(labels, Variable) else labels.contiguous().view(-1, 1)
    
    if opt.cuda:
        binary = torch.cuda.FloatTensor(batch_size, n_classes).zero_().scatter_(1, idx.long(), 1)
    else:
        binary = torch.FloatTensor(batch_size, n_classes).zero_().scatter_(1, idx.long(), 1)
    
    return Variable(binary) if isinstance(labels, Variable) else binary


# In[ ]:


class View(nn.Module):
    """
        Implements a reshaping module.
        Allows to reshape a tensor between NN layers.
    """
    
    def __init__(self, *shape):
        super(View, self).__init__()
        self.shape = shape
    
    def forward(self, input):
        return input.view(self.shape)


# In[ ]:


class LogisticRegression(nn.Module):
    def __init__(self, input_size, num_classes):
        super(LogisticRegression, self).__init__()
        self.linear = nn.Linear(input_size, num_classes)
    
    def forward(self, x):
        out = self.linear(x.view(-1, 784))
        return out


# In[ ]:


import torch.nn.functional as F

filterSize = 5
w_out = 4
h_out = 4

class SimpleConvNet(nn.Module):
    
    def __init__(self, ngpu):
        super(SimpleConvNet, self).__init__()
        self.ngpu = ngpu
        
        self.main = nn.Sequential(
            nn.Conv2d(opt.nChannels, opt.ndf, filterSize),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(2, 2),
            nn.BatchNorm2d(opt.ndf),
            nn.Conv2d(opt.ndf, opt.ndf, filterSize),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(opt.ndf),
            nn.MaxPool2d(2, 2),
            View(-1, opt.ndf * w_out * h_out),
            nn.Linear(opt.ndf * w_out * h_out, 384),
            nn.ReLU(inplace=True),
            nn.Linear(384, 384),
            nn.ReLU(inplace=True),
            nn.Linear(384, opt.nClasses),
        )

    def forward(self, x):
        return self.main(x)
    
    @staticmethod
    def create(ngpu):
        """
            Static initialisation method.
            
            Returns:
                Initialised discriminator instance.
        """
        netD = SimpleConvNet(ngpu)
        netD.apply(weights_init)
        return netD


# In[ ]:


def validate(teacher, student, valset=None, val_batches=10):
    """
        Compute test accuracy.
    """
    correct = 0.0
    total = 0.0
    
    noise = torch.FloatTensor(val_batches, opt.batchSize, nz)
    labels = torch.LongTensor(val_batches, opt.batchSize, 1)
    labels_logits = torch.FloatTensor(opt.batchSize, opt.nClasses + 1)
    
    if opt.cuda:
        noise = noise.cuda(gpu_id)
        labels = labels.cuda(gpu_id)
        labels_logits = labels_logits.cuda(gpu_id)
    
    noise.normal_(0, 1)
    labels.random_(to=opt.nClasses)
    
    fake_loss = 0
    for noise_batch, labels_batch in zip(noise, labels):
        noisev, labelv = Variable(noise_batch), Variable(labels_batch.view(-1) % 10)
        images = teacher(noisev, labelv)
        
        outputs = student(images)
        fake_loss += torch.nn.functional.cross_entropy(outputs, labelv).data.cpu().numpy()
    fake_loss = fake_loss / val_batches
    
    real_loss = 0
    real_val_batches = len(valset.dataset) / opt.batchSize + 1
    if valset is not None:
        real_loss = 0
        for images, labels in valset:
            if opt.cuda:
                images = images.cuda(gpu_id)
                labels = labels.cuda(gpu_id)
        
            images, labelv = Variable(images), Variable(labels)
            
            outputs = student(images)
            real_loss += torch.nn.functional.cross_entropy(outputs, labelv).data.cpu().numpy()
        real_loss = real_loss / real_val_batches
    
    print('Artificial validation loss: %f' % fake_loss)
    print('Real validation loss: %f' % real_loss)
    return fake_loss, real_loss


# In[ ]:


def test(testloader, net):
    """
        Compute test accuracy.
    """
    correct = 0.0
    total = 0.0
    
    '''
    if opt.cuda:
        net = net.cuda()
    '''
    
    for data in testloader:
        images, labels = data
        
        if opt.cuda:
            images = images.cuda(gpu_id)
            labels = labels.cuda(gpu_id)
            
        outputs = net(Variable(images))
        _, predicted = torch.max(outputs.data, 1)
        #print(predicted.cpu().numpy())
        #print(labels.cpu().numpy())
        total += labels.size(0)
        correct += (predicted == (labels.long().view(-1) % 10)).sum()
        #print torch.cat([predicted.view(-1, 1), (labels.long() % 10)], dim=1)

    print('Accuracy of the network on test images: %f %%' % (100 * float(correct) / total))
    return 100 * float(correct) / total


# In[ ]:


from scipy.spatial.distance import pdist
from scipy.stats import t, bayes_mvs, binom
from scipy.special import logsumexp

class BayesianAccountant:
    ''' Bayesian privacy accountant for Gaussian noise mechanism '''
    
    def __init__(self, power=32, conf=1-1e-16):
        self.power = power
        self.conf = conf
        self.logmgf = 0
        self.logconf = 0
        self.history = []
        
    def log_binom_expect(self, log_func, args=()):
        d, n, p = args
        k = np.arange(n+1)
        log_binom_coefs = binom.logpmf(k, n=n, p=p)
        return logsumexp(log_binom_coefs + log_func(d, k))

    def accumulate(self, grads, sigma, q=1, steps=1):
        if np.isscalar(grads):
            log_gauss_ratio = lambda d, k: d**2 * k * (k - 1) / (2 * sigma**2)
            self.history += [self.logmgf]
            self.logmgf += steps * self.log_binom_expect(log_gauss_ratio, args=(grads, self.power + 1, q))
            self.logconf += 0
            return self.logmgf, self.logconf
        
        dists = pdist(grads)
        n_samples = len(dists)
        
        log_gauss_ratio = lambda d, k: d**2 * k * (k - 1) / (2 * sigma**2)
        log_mgf_all = [steps * self.log_binom_expect(log_gauss_ratio, args=(d, self.power + 1, q)) for d in dists]
        log_mgf_all = np.array(log_mgf_all)

        mgf_mu = np.mean(np.exp(log_mgf_all))
        mgf_std = np.std(np.exp(log_mgf_all))

        self.history += [self.logmgf]
        self.logmgf += np.log(mgf_mu + t.ppf(q=self.conf, df=n_samples-1) * mgf_std / np.sqrt(n_samples - 1))
        self.logconf += np.log(self.conf)

        return self.logmgf, self.logconf
    
    def get_privacy(self, target_eps=None, target_delta=1e-5):
        if target_eps or (target_delta is None):
            raise ValueError("Computing delta from epsilon is not surrorted yet.")
        return (self.logmgf - np.log(target_delta - (1 - np.exp(self.logconf)))) / self.power


# In[ ]:


def sparsify_update(params, p, use_grad_field=True):
    init = True
    for param in params:
        if param is not None:
            if init:
                idx = torch.zeros_like(param, dtype=torch.bool)
                idx.bernoulli_(1 - p)
            if use_grad_field:
                if param.grad is not None:
                    idx = torch.zeros_like(param, dtype=torch.bool)
                    idx.bernoulli_(1 - p)
                    param.grad.data[idx] = 0
            else:
                init = False
                param.data[idx] = 0
    return idx


# In[ ]:


from joblib import Parallel, delayed

def train(trainloader, student, n_epoch=25, lr=0.0001, accountant=None):
    criterion = nn.CrossEntropyLoss(reduction='none')
    optimizer = optim.Adam(student.parameters(), lr=lr)
    #scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=3, gamma=0.1)
    
    if opt.cuda:
        student = student.cuda(gpu_id)
        criterion = criterion.cuda(gpu_id)
    
    accuracies = []
    
    num_batches = len(trainloader.dataset) / opt.batchSize + 1
    sampling_prob = 0.25
    num_samples = 1
    #q = float(opt.batchSize) / len(trainloader.dataset)
    max_grad_norm = opt.C
    
    for epoch in range(n_epoch):  # loop over the dataset multiple times
        running_loss = 0.0
        for i, data in enumerate(trainloader, 0):
            # get the inputs
            inputs, labels = data
            
            if opt.cuda:
                inputs = inputs.cuda(gpu_id)
                labels = labels.cuda(gpu_id)
            
            inputv = Variable(inputs)
            labelv = Variable(labels.long().view(-1) % 10)
            
            batch_size = float(len(inputs))
            
            # zero the parameter gradients
            optimizer.zero_grad()
            
            # forward + backward + optimize
            outputs = student(inputv)
            loss = criterion(outputs, labelv)
            
            if accountant:
                grads_est = []
                num_subbatch = 8
                for j in range(num_subbatch):
                    grad_sample = torch.autograd.grad(loss[np.delete(range(int(batch_size)), j)].mean(), [p for p in student.parameters() if p.requires_grad], retain_graph=True)
                    with torch.no_grad():
                        grad_sample = torch.cat([g.view(-1) for g in grad_sample])
                        grad_sample /= max(1.0, grad_sample.norm().item() / max_grad_norm)
                        grads_est += [grad_sample]
                with torch.no_grad():
                    grads_est = torch.stack(grads_est)
                    sparsify_update(grads_est, p=sampling_prob, use_grad_field=False)
                
            (loss.mean()).backward()
            
            if accountant:
                torch.nn.utils.clip_grad_norm_(student.parameters(), max_grad_norm)
                for group in optimizer.param_groups:
                    for p in group['params']:
                        if p.grad is not None:
                            p.grad += torch.randn_like(p.grad) * (opt.sigma * max_grad_norm)
                sparsify_update(student.parameters(), p=sampling_prob)
                
            optimizer.step()
            
            if accountant:
                batch_size = float(len(inputs))
                q = batch_size / len(trainloader.dataset)
                accountant.accumulate(grads=grads_est.cpu(), sigma=opt.sigma*max_grad_norm, q=q, steps=num_samples)
                running_eps = accountant.get_privacy(target_delta=1e-5)
            else:
                running_eps = -1
            
            i += 1
            # print statistics
            running_loss += loss.data[0]
            if i % 50 == 0:    # print every 100 mini-batches
                print('[%d, %5d] loss: %.3f; privacy: %f' %
                      (epoch + 1, i, running_loss / 100, running_eps))
                running_loss = 0.0
                
        #scheduler.step()
        acc = test(testloader, student)
        accuracies += [acc]
        print("Test accuracy is %d %%" % acc)

    torch.save(student.state_dict(), '%s/netL_epoch_%d.pth' % (opt.outf, n_epoch))
    print('Finished Training')
    return student.cpu(), accuracies


# In[ ]:


# transformations applied to data
transform = transforms.Compose([transforms.ToTensor()])

# switch datasets
if opt.dataset == 'adult':
    train_file = opt.dataroot + os.sep + opt.dataset + os.sep + "adult.data"
    test_file = opt.dataroot + os.sep + opt.dataset + os.sep + "adult.test"
    trainset = read_census_dataset(train_file)
    testset = read_census_dataset(test_file)
elif opt.dataset == 'mnist':
    trainset = torchvision.datasets.MNIST(root=opt.dataroot + os.sep + opt.dataset, train=True, download=True, transform=transform)
    testset = torchvision.datasets.MNIST(root=opt.dataroot + os.sep + opt.dataset, train=False, download=True, transform=transform)
elif opt.dataset == 'cifar10':
    trainset = torchvision.datasets.CIFAR10(root=opt.dataroot + os.sep + opt.dataset, train=True, download=True, transform=transform)
    testset = torchvision.datasets.CIFAR10(root=opt.dataroot + os.sep + opt.dataset, train=False, download=True, transform=transform)
elif opt.dataset == 'svhn':
    trainset = torchvision.datasets.SVHN(root=opt.dataroot + os.sep + opt.dataset, split='train', download=True, transform=transform)
    testset = torchvision.datasets.SVHN(root=opt.dataroot + os.sep + opt.dataset, split='test', download=True, transform=transform)

#val_size = 10000
#valset = torch.utils.data.Subset(trainset, range(val_size))
#trainset = torch.utils.data.Subset(trainset, range(val_size, len(trainset)))
    
# initialise data loaders
trainloader = torch.utils.data.DataLoader(trainset, batch_size=opt.batchSize, shuffle=True, num_workers=2, drop_last=True)
#valloader = torch.utils.data.DataLoader(valset, batch_size=opt.batchSize, shuffle=True, num_workers=2)
testloader = torch.utils.data.DataLoader(testset, batch_size=opt.batchSize, shuffle=True, num_workers=2)

# names of classes
classes = tuple(np.arange(10).astype(str))


# In[ ]:


# train GAN and measure time
start = time.time()

netL = SimpleConvNet(ngpu)

bayes_accountant = BayesianAccountant(power=32)
netL, accs = train(trainloader, netL, lr=0.0005, n_epoch=20, accountant=bayes_accountant)

stop = time.time()
print("Time elapsed: %f" % (stop - start))


# In[ ]:


bayes_accountant.get_privacy(target_delta=1e-10)

