
# coding: utf-8

# In[ ]:


# Generic imports
import argparse
import numpy as np
import pandas as pd
import os
import pickle
import random
import time
import matplotlib.pyplot as plt
get_ipython().run_line_magic('matplotlib', 'inline')

# PyTorch imports
import torch
import torchvision
import torchvision.transforms as transforms
import torch.nn as nn
import torch.nn.functional as func
import torch.optim as optim
import torch.backends.cudnn as cudnn
from torch.optim.optimizer import required
from torch.autograd import Variable
from torch.autograd import Function

# Other stuff
from bayesian_accountant import BayesianAccountant
from private_gan_training import BayesianPrivacyGenerativePlatform
from student_training import TeachingPlatform


# In[ ]:


# Arguments and setup
parser = argparse.ArgumentParser()

# Data loader and output
parser.add_argument('--dataset', default='fashion-mnist', help='mnist | cifar10 | svhn')
parser.add_argument('--dataroot', default='data', help='path to dataset')
parser.add_argument('--workers', type=int, default=2, help='number of data loading workers')
parser.add_argument('--batchSize', type=int, default=512, help='input batch size')
parser.add_argument('--outf', default='output_fashion_mnist', help='folder to output images and model checkpoints')
parser.add_argument('--saveImages', type=bool, default=True, help='save generated images. default=True')

# Data properties
parser.add_argument('--imageSize', type=int, default=28, help='the height / width of the input image to network')
parser.add_argument('--nChannels', type=int, default=1, help='number of colour channels')
parser.add_argument('--nClasses', type=int, default=10, help='number of labels (classes)')

# GAN parameters
parser.add_argument('--conditional', type=bool, default=True, help='condition GAN on class labels')
parser.add_argument('--nz', type=int, default=100, help='size of the latent z vector')
parser.add_argument('--ngf', type=int, default=64)
parser.add_argument('--ndf', type=int, default=64)
parser.add_argument('--filterSize', type=int, default=5)
parser.add_argument('--netG', default='', help="path to netG (to continue training)")
parser.add_argument('--netD', default='', help="path to netD (to continue training)")

# Learning parameters
parser.add_argument('--niter', type=int, default=300, help='number of epochs to train for')
parser.add_argument('--lr', type=float, default=0.0002, help='learning rate, default=0.0002')
parser.add_argument('--nCriticIter', type=int, default=5, help='number of critic iterations')
parser.add_argument('--nGeneratorIter', type=int, default=1, help='number of generator iterations')
parser.add_argument('--beta1', type=float, default=0.0, help='beta1 for adam. default=0.5')
parser.add_argument('--beta2', type=float, default=0.9, help='beta2 for adam. default=0.9')
parser.add_argument('--lambda_gp', type=float, default=10, help='gradient penalty weight, default=0.5')

# Privacy parameters
parser.add_argument('--C', type=float, default=0.5, help='embedding L2-norm bound, default=1.0')
parser.add_argument('--C_G', type=float, default=None, help='embedding L2-norm bound, default=None')
parser.add_argument('--sigma', type=float, default=0.01, help='noise variance, default=0.5')

# Computing parameters
parser.add_argument('--cuda', action='store_true', help='enables cuda')
parser.add_argument('--ngpu', type=int, default=1, help='number of GPUs to use')
parser.add_argument('--gpuId', type=int, default=1, help='ID of GPU to use')
parser.add_argument('--manualSeed', type=int, help='manual seed')

opt, unknown = parser.parse_known_args()

try:
    os.makedirs(opt.outf)
except OSError:
    pass

if torch.cuda.is_available():
    print("Using CUDA")
    opt.cuda = True
    if opt.ngpu < 0:
        opt.ngpu = torch.cuda.device_count()
        opt.gpuId = range(opt.ngpu)
    
if opt.manualSeed is None:
    opt.manualSeed = random.randint(1, 10000)

print("Random Seed: ", opt.manualSeed)
random.seed(opt.manualSeed)
torch.manual_seed(opt.manualSeed)
if opt.cuda:
    torch.cuda.manual_seed_all(opt.manualSeed)

cudnn.benchmark = True


# In[ ]:


def compute_dim_out(num_conv_layers, dim_in, kernel_size=opt.filterSize, stride=2, padding=0, dilation=1):
    dim_out = dim_in
    for i in range(num_conv_layers):
        dim_out = np.int((dim_out + 2 * padding - dilation * (kernel_size - 1) - 1) / stride + 1)
    return dim_out

def onehot(labels, batch_size, n_classes):
    if not opt.conditional:
        return labels.float()
    
    idx = labels.data.contiguous().view(-1, 1) if isinstance(labels, Variable) else labels.contiguous().view(-1, 1)
    binary = torch.zeros((batch_size, n_classes), device=labels.device).scatter_(1, idx.long(), 1)
    
    return Variable(binary) if isinstance(labels, Variable) else binary


# In[ ]:


class View(nn.Module):
    """
        Implements a reshaping module.
        Allows to reshape a tensor between NN layers.
    """
    
    def __init__(self, *shape):
        super(View, self).__init__()
        self.shape = shape
    
    def forward(self, input):
        return input.view(self.shape)


# In[ ]:


def weights_init(m):
    """
        Weight initialiser
    """
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        m.weight.data.normal_(0.0, 0.02)
    elif classname.find('BatchNorm') != -1:
        m.weight.data.normal_(1.0, 0.02)
        m.bias.data.fill_(0)
        

class G(nn.Module):
    """
        Generator class.
    """
    
    def __init__(self, conditional=False):
        super(G, self).__init__()
        self.conditional = conditional
        self.n_classes = opt.nClasses if self.conditional else 0
        self.stride = 2
        
        self.main = nn.Sequential(
            # input is going into linear layer
            nn.Linear(opt.nz + self.n_classes, 256, bias=False),
            nn.SELU(inplace=True),
            
            nn.Linear(256, opt.ngf * 8 * 4 * 4, bias=False),
            nn.SELU(inplace=True),
            
            View(-1, opt.ngf * 8, 4, 4),
            
            nn.ConvTranspose2d(opt.ngf * 8, opt.ngf * 4, opt.filterSize, stride=self.stride, bias=False),
            nn.SELU(inplace=True),
            
            nn.ConvTranspose2d(opt.ngf * 4, opt.ngf * 2, opt.filterSize, stride=self.stride, bias=False),
            nn.SELU(inplace=True),
            
            # state size. (ngf*8) x 4 x 4
            nn.ConvTranspose2d(opt.ngf * 2, opt.ngf * 1, opt.filterSize, stride=self.stride, bias=False),
            nn.SELU(inplace=True),
            
            # state size. (ngf) x 32 x 32
            nn.ConvTranspose2d(opt.ngf * 1, opt.nChannels, opt.filterSize, stride=self.stride, bias=False),
            nn.AdaptiveMaxPool2d(output_size=(opt.imageSize, opt.imageSize)),
        )
        
        self.out = nn.Sequential(
            nn.Tanh(),
        )
        
        
    def forward(self, input, label=None):
        """
            Forward pass.
            
            Args:
                input: noise vector (1D).
                
            Return:
                Generated image.
        """
        if self.conditional:
            batch_size = input.size(0)
            if label is not None:
                full_input = torch.cat([input, onehot(label, batch_size, self.n_classes)], dim=1)
            else:
                full_input = torch.cat([input, torch.zeros(batch_size, self.n_classes, device=input.device)], dim=1)
        else:
            full_input = input
        
        output = self.main(full_input)
        output = self.out(output)
        return output

    
    @staticmethod
    def create(conditional=False):
        """
            Static initialisation method.
            
            Returns:
                Initialised generator instance.
        """
        netG = G(conditional)
        netG.apply(weights_init)
        if opt.netG != '':
            netG.load_state_dict(torch.load(opt.netG))
        return netG


# In[ ]:


class GaussianNoiseLayer(nn.Module):
    """
        Implements a Gaussian noise module.
        Adds Gaussian noise with mean 0 and variance \sigma to the output of the previous layer.
    """
    def __init__(self, sigma):
        super(GaussianNoiseLayer, self).__init__()
        self.sigma = sigma
    
    def forward(self, input):
        if opt.cuda:
            noise = Variable(torch.cuda.FloatTensor(input.size()).normal_(std=self.sigma))
        else:
            noise = Variable(torch.FloatTensor(input.size()).normal_(std=self.sigma))
        return input.add(noise)


class ClipNormLayer(nn.Module):
    """
        Implements a clipping module.
        Clips the norm of the output of the previous layer to a specified value.
    """
    def __init__(self, p=2, max_value=1.0):
        super(ClipNormLayer, self).__init__()
        self.p = p
        self.max = max_value
    
    def forward(self, input):
        return torch.renorm(input, p=self.p, dim=0, maxnorm=self.max)
    
    
class D(nn.Module):
    
    def __init__(self, conditional=False):
        super().__init__()
        self.conditional = conditional
        self.n_classes = opt.nClasses if self.conditional else 0
        
        self.stride = 2
        self.padding = 1
        self.h_out = compute_dim_out(3, opt.imageSize, stride=self.stride, padding=self.padding)
        self.w_out = compute_dim_out(3, opt.imageSize, stride=self.stride, padding=self.padding)
        self.embedding_dim = opt.ndf * 4 * self.h_out * self.w_out
        
        self.conv = nn.Sequential(
            # input is (nc) x 64 x 64
            nn.Conv2d(opt.nChannels, opt.ndf, opt.filterSize, stride=self.stride, padding=self.padding, bias=False),
            nn.SELU(inplace=True),
            
            # state size. (ndf) x 32 x 32
            nn.Conv2d(opt.ndf * 1, opt.ndf * 2, opt.filterSize, stride=self.stride, padding=self.padding, bias=False),
            nn.SELU(inplace=True),
            
            # state size. (ndf*2) x 16 x 16
            nn.Conv2d(opt.ndf * 2, opt.ndf * 4, opt.filterSize, stride=self.stride, padding=self.padding, bias=False),
            nn.SELU(inplace=True),
            
            #nn.Conv2d(opt.ndf * 4, opt.ndf * 8, opt.filterSize, stride=self.stride, padding=self.padding, bias=False),
            #nn.SELU(inplace=True),
        )
        
        self.classifier = nn.Sequential(
            nn.Linear(self.embedding_dim + self.n_classes, 256, bias=False),
            #ClipNormLayer(max_value=1),
            #GaussianNoiseLayer(sigma=1),
            nn.SELU(inplace=True),
            nn.Linear(256, 1, bias=False)
        )
        
    
    def forward(self, input, label=None):
        """
            Forward pass.
            
            Args:
                input: image to classify (3D tensor - colour, width, height).
                
            Return:
                Predicted image class.
        """
        
        output = self.conv(input)
        output = output.view(-1, self.embedding_dim)
        if self.conditional:
            if label is not None:
                output = torch.cat([output, onehot(label, label.size(0), self.n_classes)], dim=1)
            else:
                output = torch.cat([output, torch.zeros(output.size(0), self.n_classes, device=output.device)], dim=1)
        output = self.classifier(output)
        
        return output
    
    
    @staticmethod
    def create(conditional=False):
        """
            Static initialisation method.
            
            Returns:
                Initialised discriminator instance.
        """
        netD = D(conditional)
        netD.apply(weights_init)
        if opt.netD != '':
            netD.load_state_dict(torch.load(opt.netD))
        return netD


# In[ ]:


import torch.nn.functional as F

filterSize = 5
w_out = 5
h_out = 5

class SimpleConvNet(nn.Module):
    
    def __init__(self):
        super().__init__()
        
        self.filterSize = 5
        self.h_out = compute_dim_out(3, opt.imageSize, kernel_size=self.filterSize, padding=1)
        self.w_out = compute_dim_out(3, opt.imageSize, kernel_size=self.filterSize, padding=1)
        self.embedding_dim = opt.ndf * 4 * self.h_out * self.w_out
        
        self.main = nn.Sequential(
            nn.Conv2d(opt.nChannels, opt.ndf, self.filterSize, stride=2, padding=1, bias=False),
            nn.SELU(inplace=True),
            
            # state size. (ndf) x 32 x 32
            nn.Conv2d(opt.ndf * 1, opt.ndf * 2, self.filterSize, stride=2, padding=1, bias=False),
            nn.SELU(inplace=True),
            
            # state size. (ndf*2) x 16 x 16
            nn.Conv2d(opt.ndf * 2, opt.ndf * 4, self.filterSize, stride=2, padding=1, bias=False),
            nn.SELU(inplace=True),
            
            View(-1, self.embedding_dim),
            nn.Linear(self.embedding_dim, 64),
            nn.SELU(inplace=True),
            nn.Linear(64, opt.nClasses),
        )

    def forward(self, x):
        return self.main(x)


# In[ ]:


# transformations applied to data
transform = transforms.Compose([transforms.ToTensor()])

# switch datasets
if opt.dataset == 'adult':
    train_file = opt.dataroot + os.sep + opt.dataset + os.sep + "adult.data"
    test_file = opt.dataroot + os.sep + opt.dataset + os.sep + "adult.test"
    trainset = read_census_dataset(train_file)
    testset = read_census_dataset(test_file)
elif opt.dataset == 'mnist':
    trainset = torchvision.datasets.MNIST(root=opt.dataroot + os.sep + opt.dataset, train=True, download=True, transform=transform)
    testset = torchvision.datasets.MNIST(root=opt.dataroot + os.sep + opt.dataset, train=False, download=True, transform=transform)
elif opt.dataset == 'cifar10':
    trainset = torchvision.datasets.CIFAR10(root=opt.dataroot + os.sep + opt.dataset, train=True, download=True, transform=transform)
    testset = torchvision.datasets.CIFAR10(root=opt.dataroot + os.sep + opt.dataset, train=False, download=True, transform=transform)
elif opt.dataset == 'svhn':
    trainset = torchvision.datasets.SVHN(root=opt.dataroot + os.sep + opt.dataset, split='train', download=True, transform=transform)
    testset = torchvision.datasets.SVHN(root=opt.dataroot + os.sep + opt.dataset, split='test', download=True, transform=transform)
elif opt.dataset == 'fashion-mnist':
    trainset = torchvision.datasets.FashionMNIST(root=opt.dataroot + os.sep + opt.dataset, train=True, download=True, transform=transform)
    testset = torchvision.datasets.FashionMNIST(root=opt.dataroot + os.sep + opt.dataset, train=False, download=True, transform=transform)

#val_size = 10000
#valset = torch.utils.data.Subset(trainset, range(val_size))
#trainset = torch.utils.data.Subset(trainset, range(val_size, len(trainset)))
    
# initialise data loaders
trainloader = torch.utils.data.DataLoader(trainset, batch_size=opt.batchSize, shuffle=True, num_workers=2, drop_last=True)
#valloader = torch.utils.data.DataLoader(valset, batch_size=opt.batchSize, shuffle=True, num_workers=2)
testloader = torch.utils.data.DataLoader(testset, batch_size=opt.batchSize, shuffle=True, num_workers=2)

# names of classes
classes = tuple(np.arange(10).astype(str))


# In[ ]:


# train GAN and measure time
start = time.time()

bayes_accountant = BayesianAccountant(power=48, exchangeable=True)
platform = BayesianPrivacyGenerativePlatform(accountant=bayes_accountant, options=opt)

netG = G.create(conditional=True)
netD = D.create(conditional=True)

# start traing from later epoch
epoch_start = 0
if epoch_start > 0:
    netG.load_state_dict(torch.load("%s/netG_epoch_%d.pth" % (opt.outf, epoch_start-1)))
    netD.load_state_dict(torch.load("%s/netD_epoch_%d.pth" % (opt.outf, epoch_start-1)))

netG, netD = platform.train_gan(netG, netD, trainloader, epoch_start=epoch_start, n_epochs=201)

stop = time.time()
print("Time elapsed: %f" % (stop - start))


# In[ ]:


print("Final epsilon: ", bayes_accountant.get_privacy(target_delta=1e-10))


# In[ ]:


import torch.nn.functional as F

filterSize = 5
w_out = 4
h_out = 4

class Labeller(nn.Module):
    def __init__(self):
        super(Labeller, self).__init__()
        self.main = nn.Sequential(
            nn.Conv2d(opt.nChannels, opt.ndf, filterSize),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(2, 2),
            nn.BatchNorm2d(opt.ndf),
            nn.Conv2d(opt.ndf, opt.ndf, filterSize),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(opt.ndf),
            nn.MaxPool2d(2, 2),
            View(-1, opt.ndf * w_out * h_out),
            nn.Linear(opt.ndf * w_out * h_out, 384),
            nn.ReLU(inplace=True),
            nn.Linear(384, 384),
            nn.ReLU(inplace=True),
            nn.Linear(384, opt.nClasses),
        )
        
    def forward(self, x):
        return self.main(x)
    
    def labelled_dataset(self, generator, batch_size, total_size=None, labels=None):
        print("Generating training images...")
        with torch.no_grad():
            if labels is None:
                labels = torch.randint(0, opt.nClasses, size=(total_size, 1), dtype=torch.long)
            else:
                total_size = len(labels)
            noise = torch.randn(total_size, opt.nz)
            num_batches = int(np.ceil(total_size / batch_size))
            images = []
            for i in range(num_batches):
                noise_batch = noise[i*batch_size : (i+1)*batch_size].to(next(generator.parameters()))
                labels_batch = labels[i*batch_size : (i+1)*batch_size].to(next(generator.parameters()))
                image_batch = generator(noise_batch, labels_batch)
                images += [image_batch.cpu()]
                labels[i*batch_size : (i+1)*batch_size] = torch.argmax(self(image_batch), dim=1).unsqueeze(-1)
            images = torch.cat(images, dim=0)
            labels = labels.to(images)
        dataset = torch.utils.data.TensorDataset(images, labels)
        dataloader = torch.utils.data.DataLoader(dataset, batch_size=opt.batchSize, shuffle=True)
        return dataloader

teacher = G(conditional=True)
teacher.load_state_dict(torch.load('%s/netG_epoch_%d.pth' % (opt.outf, 200)))
teacher.cuda(opt.gpuId)
    
netL = Labeller()
netL.load_state_dict(torch.load("output_fashion_mnist_labeller/netL_epoch_%d.pth" % (50)))
netL.cuda(opt.gpuId)

privateloader = netL.labelled_dataset(teacher, opt.batchSize, total_size=60000)


# In[ ]:


netL.cuda(opt.gpuId)
netG.cuda(opt.gpuId)    
sizes = [100, 200, 500, 1000, 2000, 5000, 10000, 60000]
accs_per_size = []
for s in sizes:
    accs_per_iter = []
    for _ in range(10):
        net = SimpleConvNet()
        privateloader = netL.labelled_dataset(netG, opt.batchSize, total_size=s)
        teaching_platform = TeachingPlatform(options=opt)
        net, accs = teaching_platform.teach(net, privateloader, testloader, n_epoch=50, lr=0.0001)
        accs_per_iter += [accs[-1]]
    accs_per_size += [accs_per_iter]
pickle.dump(accs_per_size, open('accs_less_labels_fashion_mnist_bdp.pickle', 'wb'))


# In[ ]:


netL.cuda(opt.gpuId)
netG.cuda(opt.gpuId)  
accs_per_size_np = []
for s in sizes:
    accs_per_iter = []
    for _ in range(10):
        net = SimpleConvNet()
        privateset = torch.utils.data.Subset(trainset, np.random.choice(60000, size=s, replace=False))
        privateloader = torch.utils.data.DataLoader(privateset, batch_size=opt.batchSize, shuffle=True)
        teaching_platform = TeachingPlatform(options=opt)
        net, accs = teaching_platform.teach(net, privateloader, testloader, n_epoch=50, lr=0.0001)
        accs_per_iter += [accs[-1]]
    accs_per_size_np += [accs_per_iter]
pickle.dump(accs_per_size_np, open('accs_less_labels_fashion_mnist_nop.pickle', 'wb'))


# In[ ]:


y_mean_bp = np.mean(accs_per_size, axis=1) / np.array(np.mean(accs_per_size, axis=1)).max() * 100
y_err_bp = 2 * np.std(accs_per_size, axis=1)
y_mean_np = np.mean(accs_per_size_np, axis=1) / np.array(np.mean(accs_per_size_np, axis=1)).max() * 100
y_err_np = 2 * np.std(accs_per_size_np, axis=1)

plt.errorbar(sizes, y_mean_bp, yerr=y_err_bp, fmt='C2-', label='Synthetic')
plt.fill_between(sizes, y_mean_bp - y_err_bp, y_mean_bp + y_err_bp, color='C2', alpha=0.3)
plt.errorbar(sizes, y_mean_np, yerr=y_err_np, fmt='C3--', label='Real')
plt.fill_between(sizes, y_mean_np - y_err_np, y_mean_np + y_err_np, color='C3', alpha=0.3)
plt.legend()
plt.xscale('log')
#plt.yscale('log')
plt.xlabel('Labelled images')
plt.ylabel('Test accuracy, % of maximum')
plt.savefig('fashion_mnist_less_labels.pdf', format='pdf', bbox_inches='tight')


# In[ ]:


print("Test accuracy (mean): ", np.mean(accs_per_size, axis=1))

