import numpy as np

from scipy.spatial.distance import pdist
from scipy.stats import t, bayes_mvs, binom
from scipy.special import logsumexp

class BayesianAccountant:
    ''' Bayesian privacy accountant for Gaussian noise mechanism '''
    
    def __init__(self, power=32, conf=1-1e-16, exchangeable=True):
        self.power = power
        self.conf = conf
        self.exchangeable = exchangeable
        self.logmgf = 0
        self.logconf = 0
        self.history = []
        
    def log_binom_expect(self, log_func, args=()):
        d, n, p = args
        k = np.arange(n+1)
        log_binom_coefs = binom.logpmf(k, n=n, p=p)
        return logsumexp(log_binom_coefs + log_func(d, k), axis=1)

    def accumulate(self, grads, sigma, q=1, steps=1):
        
        if np.isscalar(grads):
            log_gauss_ratio = lambda d, k: d**2 * k * (k - 1) / (2 * sigma**2)
            self.history += [self.logmgf]
            self.logmgf += steps * self.log_binom_expect(log_gauss_ratio, args=(np.array([[grads]]), self.power + 1, q))
            self.logconf += 0
            return self.logmgf, self.logconf
        
        dists = pdist(grads)
        n_samples = len(dists)
        
        log_gauss_ratio = lambda d, k: np.outer(d**2, k * (k - 1) / (2 * sigma**2))
        log_mgf_all = steps * self.log_binom_expect(log_gauss_ratio, args=(dists, self.power + 1, q))
        log_mgf_all = np.array(log_mgf_all)

        mgf_mu = np.mean(np.exp(log_mgf_all))
        mgf_std = np.std(np.exp(log_mgf_all))
        
        # Numerically (more) stable version:
        #     mgf_mu = np.exp(-np.log(n_samples) + logsumexp(log_mgf_all))
        #     mgf_std = np.sqrt(np.exp(-np.log(n_samples) + logsumexp(2 * log_mgf_all)) - mgf_mu**2)

        self.history += [self.logmgf]
        self.logmgf += np.log(mgf_mu + t.ppf(q=self.conf, df=n_samples-1) * mgf_std / np.sqrt(n_samples - 1))
        self.logconf += np.log(self.conf)

        return self.logmgf, self.logconf
    
    def get_privacy(self, target_eps=None, target_delta=1e-5):
        if target_eps or (target_delta is None):
            raise ValueError("Computing delta from epsilon is not surrorted yet.")
        correction = 1.0 if self.exchangeable else (1.0 / (len(self.history) + 1))
        return (correction * self.logmgf - np.log(target_delta - (1 - np.exp(self.logconf)))) / self.power
    
    def set_sensitivity(self, sensitivity):
        self.sensitivity = sensitivity
    
    def set_noise_std(self, sigma):
        self.sigma = sigma
    
    def get_sensitivity(self):
        return self.sensitivity
    
    def get_noise_std(self):
        return self.sigma