# Generic imports
import argparse
import numpy as np
import pandas as pd
import os
import pickle
import random
import time

# PyTorch imports
import torch
import torchvision
import torchvision.transforms as transforms
import torch.nn as nn
import torch.nn.functional as func
import torch.optim as optim
import torch.backends.cudnn as cudnn
from torch.optim.optimizer import required
from torch.autograd import Variable
from torch.autograd import Function


def sparsify_update(params, p, use_grad_field=True):
    init = True
    for param in params:
        if param is not None:
            if init:
                idx = torch.zeros_like(param, dtype=torch.bool)
                idx.bernoulli_(1 - p)
            if use_grad_field:
                if param.grad is not None:
                    idx = torch.zeros_like(param, dtype=torch.bool)
                    idx.bernoulli_(1 - p)
                    param.grad.data[idx] = 0
            else:
                init = False
                param.data[idx] = 0
    return idx


class BayesianPrivacyGenerativePlatform():

    def __init__(self, accountant=None, options=None):
        self.accountant = accountant
        self.options = options


    def train_gan(self, netG, netD, trainloader, epoch_start=0, n_epochs=None):
        
        if n_epochs is None:
            n_epochs = self.options.niter
        
        optimizerG = optim.Adam(netG.parameters(), lr=self.options.lr, betas=(self.options.beta1, self.options.beta2))
        optimizerD = optim.Adam(netD.parameters(), lr=self.options.lr, betas=(self.options.beta1, self.options.beta2))
        
        noise = torch.FloatTensor(self.options.batchSize, self.options.nz)
        fixed_label = torch.FloatTensor(64, 1).random_(to=self.options.nClasses) if self.options.conditional else None
        fixed_noise = torch.FloatTensor(64, self.options.nz).normal_(0, 1)
        
        print("Generating test labels:\n", fixed_label.view(8, 8).cpu().numpy())
        
        if self.options.cuda:
            netD.cuda(self.options.gpuId)
            netG.cuda(self.options.gpuId)
            noise = noise.cuda(self.options.gpuId)
            fixed_noise = fixed_noise.cuda(self.options.gpuId)
            fixed_label = fixed_label.cuda(self.options.gpuId) if self.options.conditional else None
        
        '''
        if not isinstance(netD, nn.DataParallel):
            netD = nn.DataParallel(netD, device_ids=range(self.options.ngpu))
        if not isinstance(netG, nn.DataParallel):
            netG = nn.DataParallel(netG, device_ids=range(self.options.ngpu))
        '''
        
        schedulerG = optim.lr_scheduler.StepLR(optimizerG, step_size=1, gamma=0.9)
        schedulerD = optim.lr_scheduler.StepLR(optimizerD, step_size=1, gamma=0.9)
        
        running_eps = 0
        for epoch in range(epoch_start, epoch_start + n_epochs):
            data_stream = enumerate(trainloader)
            
            #schedulerG.step()
            #schedulerD.step()
            i = 0
            critic_iter = 0
            
            endOfEpoch = False
            endOfLabelled = False
            for inputs, labels in trainloader:
                i += 1
                critic_iter += 1
                
                # prepare data
                if self.options.cuda:
                    inputs = inputs.cuda(self.options.gpuId)
                    labels = labels.cuda(self.options.gpuId)#.view(-1) % self.options.nClasses
                
                lossD = self.update_D(netD, netG, optimizerD, inputs, labels, noise, trainloader)
                
                if critic_iter < self.options.nCriticIter:
                    continue
                
                critic_iter = 0
                
                for _ in range(self.options.nGeneratorIter):
                    lossG = self.update_G(netG, netD, optimizerG, optimizerD, inputs, labels, noise)
                
                if self.accountant:
                    running_eps = self.accountant.get_privacy(target_delta=1e-5)
                else:
                    running_eps = -1
                
                if i % 10 < self.options.nCriticIter:
                    print('[%d/%d][%d/%d] Loss_D: %.4f Loss_G: %.4f Privacy: %f' #D(x): %.4f D(G(z)): %.4f / %.4f'
                        % (epoch, epoch_start + n_epochs, i, len(trainloader), lossD, lossG, running_eps))
                if i % 100 < self.options.nCriticIter and self.options.saveImages:
                    torchvision.utils.save_image(inputs, '%s/real_samples.png' % self.options.outf, normalize=True)
                    fake = netG(fixed_noise, fixed_label)
                    torchvision.utils.save_image(fake.data, '%s/fake_samples_epoch_%03d.png' % (self.options.outf, epoch), normalize=True)
                    #print ('Privacy loss: %.5f' % accountant.get_privacy_spending()[0])
                    
            # do checkpointing
            if epoch % 1 == 0:
                torch.save(netG.state_dict(), '%s/netG_epoch_%d.pth' % (self.options.outf, epoch))
                torch.save(netD.state_dict(), '%s/netD_epoch_%d.pth' % (self.options.outf, epoch))
        
        return netG, netD


    def update_D(self, netD, netG, optimizerD, inputs, labels, noisev, trainloader):
        """
            Update parameters of the discriminator.
        """
        # set gradients to zero
        netD.zero_grad()
        critic_loss = 0

        # train with real
        D_real = netD(inputs, labels)
        
        batch_size = float(len(inputs))
        sampling_prob = 1.0
        if self.accountant:
            grads_est = []
            num_subbatch = 8
            for j in range(num_subbatch):
                grad_sample = torch.autograd.grad(-D_real[np.delete(range(int(batch_size)), j)].mean(), [p for p in netD.parameters() if p.requires_grad], retain_graph=True)
                with torch.no_grad():
                    grad_sample = torch.cat([g.view(-1) for g in grad_sample])
                    grad_sample /= max(1.0, grad_sample.norm().item() / self.options.C)
                    grads_est += [grad_sample]
            with torch.no_grad():
                grads_est = torch.stack(grads_est)
                sparsify_update(grads_est, p=sampling_prob, use_grad_field=False)
                #print(grads_est.norm(dim=1))
                        
        (-D_real.mean()).backward()
        
        if self.accountant:
            batch_size = float(len(inputs))
            q = batch_size / len(trainloader.dataset)
            self.accountant.accumulate(grads=grads_est.cpu(), sigma=self.options.sigma*self.options.C, q=q, steps=1)
            running_eps = self.accountant.get_privacy(target_delta=1e-5)
        else:
            running_eps = -1
        
        # train with fake
        noisev.data.resize_(inputs.size(0), self.options.nz).normal_(0, 1)
        fake_inputv = netG(noisev, labels)
        
        D_fake = netD(fake_inputv.detach(), labels)
        D_fake.mean().backward()

        # add gradient penalty
        if (self.options.lambda_gp) and (self.options.lambda_gp != 0):
            gp = self.gradient_penalty(netD, inputs, fake_inputv, labels)
            gp.backward()
        
        if self.accountant:
            torch.nn.utils.clip_grad_norm_(netD.parameters(), self.options.C)
            for group in optimizerD.param_groups:
                for p in group['params']:
                    if p.grad is not None:
                        p.grad += torch.randn_like(p.grad) * (self.options.sigma * self.options.C)
            sparsify_update(netD.parameters(), p=sampling_prob)
        
        optimizerD.step()
        
        # return total discriminator loss
        return (D_fake.mean() - D_real.mean()).item()


    def update_G(self, netG, netD, optimizerG, optimizerD, inputv, labelv, noisev):
        """
            Update parameters of the generator.
        """
        # set gradients to zero
        netG.zero_grad()
        gen_loss = 0
        
        noisev.data.resize_(inputv.size(0), self.options.nz).normal_(0, 1)
        
        # generate fake samples
        fake_inputv = netG(noisev, labelv)

        # train with fake
        D_fake = netD(fake_inputv, labelv)
        (-D_fake.mean()).backward()

        # TODO: it might be better not to clip/sparsify generator
        #sampling_prob = 1.0
        if self.accountant and self.options.C_G:
            torch.nn.utils.clip_grad_norm_(netG.parameters(), self.options.C_G)
            '''
            for group in optimizerG.param_groups:
                for p in group['params']:
                    if p.grad is not None:
                        p.grad += torch.randn_like(p.grad) * (self.options.sigma * self.options.C)
            '''
            #sparsify_update(netG.parameters(), p=sampling_prob)

        optimizerG.step()
        
        # return total generator loss
        return -D_fake.mean().item()

    
    def gradient_penalty(self, netD, real_input, fake_input, labelv):
        """
            Compute gradient penalty for Improved Wasserstein GAN
        """
        
        batch_size = real_input.size(0)

        eps = torch.rand(batch_size, 1, device=real_input.device)
            
        eps = eps.expand(batch_size, self.options.nChannels * self.options.imageSize * self.options.imageSize)\
                .contiguous()\
                .view(batch_size, self.options.nChannels, self.options.imageSize, self.options.imageSize)

        mixed_sample = eps * real_input.data + (1 - eps) * fake_input.data
        mixed_sample = Variable(mixed_sample, requires_grad=True)

        D_mixed = netD(mixed_sample, labelv)
        
        grad_outputs = torch.ones_like(D_mixed)

        gradients = torch.autograd.grad(outputs=D_mixed,
                                        inputs=mixed_sample,
                                        grad_outputs=grad_outputs,
                                        create_graph=True,
                                        retain_graph=True,
                                        only_inputs=True)[0]
        gradient_penalty = self.options.lambda_gp * ((gradients.view(gradients.size(0), -1).norm(2, dim=1) - 1) ** 2).mean()
        
        return gradient_penalty