# Generic imports
import argparse
import numpy as np
import pandas as pd
import os
import pickle
import random
import time

# PyTorch imports
import torch
import torchvision
import torchvision.transforms as transforms
import torch.nn as nn
import torch.nn.functional as func
import torch.optim as optim
import torch.backends.cudnn as cudnn
from torch.optim.optimizer import required
from torch.autograd import Variable
from torch.autograd import Function


def sparsify_update(params, p, use_grad_field=True):
    init = True
    for param in params:
        if param is not None:
            if init:
                idx = torch.zeros_like(param, dtype=torch.uint8)
                idx.bernoulli_(1 - p)
            if use_grad_field:
                if param.grad is not None:
                    idx = torch.zeros_like(param, dtype=torch.uint8)
                    idx.bernoulli_(1 - p)
                    param.grad.data[idx] = 0
            else:
                init = False
                param.data[idx] = 0
    return idx


class TeachingPlatform():

    def __init__(self, options=None, teachset_size=None, teaching_labels_distribution=None):
        self.options = options
        self.teachset_size = teachset_size
        self.teaching_labels_distribution = teaching_labels_distribution

    def get_dataset(self, generator, batch_size, total_size=None, labels=None):
        print("Generating training images...")
        with torch.no_grad():
            if labels is None:
                if self.teaching_labels_distribution is None:
                    labels = torch.randint(0, self.options.nClasses, size=(total_size, 1), dtype=torch.long)
                else:
                    labels = torch.multinomial(self.teaching_labels_distribution, total_size, replacement=True).unsqueeze(-1)
            else:
                total_size = len(labels)
            noise = torch.randn(total_size, self.options.nz)
            num_batches = int(np.ceil(total_size / batch_size))
            images = []
            for i in range(num_batches):
                noise_batch = noise[i*batch_size : (i+1)*batch_size].to(next(generator.parameters()))
                labels_batch = labels[i*batch_size : (i+1)*batch_size].to(next(generator.parameters()))
                images += [generator(noise_batch, labels_batch).cpu()]
            images = torch.cat(images, dim=0)
            labels = labels.to(images)
        dataset = torch.utils.data.TensorDataset(images, labels)
        dataloader = torch.utils.data.DataLoader(dataset, batch_size=self.options.batchSize, shuffle=True)
        return dataloader


    def teach(self, net, trainloader, testloader, criterion=None, optimiser=None, n_epoch=25, lr=0.0001, accountant=None, teacher=None):
        criterion = criterion if criterion else nn.CrossEntropyLoss(reduction='none')
        optimiser = optimiser if optimiser else optim.Adam(net.parameters(), lr=lr)
        #scheduler = optim.lr_scheduler.StepLR(optimiser, step_size=3, gamma=0.1)
        
        if self.options.cuda:
            net.cuda(self.options.gpuId)
            criterion.cuda(self.options.gpuId)
            if teacher:
                teacher.cuda(self.options.gpuId)

        total_size = self.teachset_size if self.teachset_size else len(trainloader.dataset)
        dataloader = self.get_dataset(teacher, self.options.batchSize, total_size) if teacher else trainloader
            
        accuracies = []
        
        num_batches = len(dataloader.dataset) // self.options.batchSize
        sampling_prob = 1.0
        #num_samples = n_epoch * num_batches
        num_samples = 1
        #q = float(self.options.batchSize) / len(dataloader.dataset)
        max_grad_norm = self.options.C

        for epoch in range(n_epoch):
            running_loss = 0.0
            i = 0
            for inputs, labels in dataloader:
                
                if self.options.cuda:
                    inputs = inputs.cuda(self.options.gpuId)
                    labels = labels.cuda(self.options.gpuId)
                # make sure labels are [0..K], where K is the number of classes
                labels = labels.long().view(-1) % self.options.nClasses

                batch_size = float(len(inputs))
                
                # zero the parameter gradients
                optimiser.zero_grad()
                
                # forward + backward + optimize
                outputs = net(inputs)
                loss = criterion(outputs, labels)
                
                if accountant:
                    grads_est = []
                    num_subbatch = 8
                    for j in range(num_subbatch):
                        grad_sample = torch.autograd.grad(loss[np.delete(range(int(batch_size)), j)].mean(), [p for p in net.parameters() if p.requires_grad], retain_graph=True)
                        with torch.no_grad():
                            grad_sample = torch.cat([g.view(-1) for g in grad_sample])
                            grad_sample /= max(1.0, grad_sample.norm().item() / max_grad_norm)
                            grads_est += [grad_sample]
                    with torch.no_grad():
                        grads_est = torch.stack(grads_est)
                        sparsify_update(grads_est, p=sampling_prob, use_grad_field=False)
                    
                (loss.mean()).backward()
                
                if accountant:
                    torch.nn.utils.clip_grad_norm_(net.parameters(), max_grad_norm)
                    for group in optimiser.param_groups:
                        for p in group['params']:
                            if p.grad is not None:
                                p.grad += torch.randn_like(p.grad) * (self.options.sigma * max_grad_norm)
                    sparsify_update(net.parameters(), p=sampling_prob)
                    
                optimiser.step()
                
                if accountant:
                    batch_size = float(len(inputs))
                    q = batch_size / len(dataloader.dataset)
                    accountant.accumulate(grads=grads_est.cpu(), sigma=self.options.sigma*max_grad_norm, q=q, steps=num_samples)
                    running_eps = accountant.get_privacy(target_delta=1e-5)
                else:
                    running_eps = -1
                
                i += 1
                # print statistics
                running_loss += loss.data[0]
                if i % 50 == 0:    # print every 50 mini-batches
                    print('[%d, %5d] loss: %.3f; privacy: %f' %
                        (epoch + 1, i, running_loss / 50, running_eps))
                    running_loss = 0.0
                    
            #scheduler.step()
            acc = self.test(net, testloader)
            accuracies += [acc]
            print("Test accuracy is %d %%" % acc)

        print('Finished Training')
        return net.cpu(), accuracies


    def validate(self, teacher, student, valset=None, val_batches=10):
        """
            Compute test accuracy.
        """
        correct = 0.0
        total = 0.0
        
        noise = torch.FloatTensor(val_batches, self.options.batchSize, nz)
        labels = torch.LongTensor(val_batches, self.options.batchSize, 1)
        labels_logits = torch.FloatTensor(self.options.batchSize, self.options.nClasses + 1)
        
        if self.options.cuda:
            noise = noise.cuda(gpu_id)
            labels = labels.cuda(gpu_id)
            labels_logits = labels_logits.cuda(gpu_id)
        
        noise.normal_(0, 1)
        labels.random_(to=self.options.nClasses)
        
        fake_loss = 0
        for noise_batch, labels_batch in zip(noise, labels):
            noisev, labelv = Variable(noise_batch), Variable(labels_batch.view(-1) % 10)
            images = teacher(noisev, labelv)
            
            outputs = student(images)
            fake_loss += torch.nn.functional.cross_entropy(outputs, labelv).data.cpu().numpy()
        fake_loss = fake_loss / val_batches
        
        real_loss = 0
        real_val_batches = len(valset.dataset) / self.options.batchSize + 1
        if valset is not None:
            real_loss = 0
            for images, labels in valset:
                if self.options.cuda:
                    images = images.cuda(gpu_id)
                    labels = labels.cuda(gpu_id)
            
                images, labelv = Variable(images), Variable(labels)
                
                outputs = student(images)
                real_loss += torch.nn.functional.cross_entropy(outputs, labelv).data.cpu().numpy()
            real_loss = real_loss / real_val_batches
        
        print('Artificial validation loss: %f' % fake_loss)
        print('Real validation loss: %f' % real_loss)
        return fake_loss, real_loss


    def test(self, net, testloader):
        """
            Compute test accuracy.
        """
        correct = 0.0
        total = 0.0
        
        '''
        if self.options.cuda:
            net = net.cuda()
        '''
        
        for data in testloader:
            images, labels = data
            
            if self.options.cuda:
                images = images.cuda(self.options.gpuId)
                labels = labels.cuda(self.options.gpuId)
                
            outputs = net(Variable(images))
            _, predicted = torch.max(outputs.data, 1)
            #print(predicted.cpu().numpy())
            #print(labels.cpu().numpy())
            total += labels.size(0)
            correct += (predicted == (labels.long().view(-1) % 10)).sum()
            #print torch.cat([predicted.view(-1, 1), (labels.long() % 10)], dim=1)

        print('Accuracy of the network on test images: %f %%' % (100 * float(correct) / total))
        return 100 * float(correct) / total